import requests
from .label import Label
from .utils import get_forta_api_headers, get_forta_api_url


def get_labels(dict):
    forta_api_url = get_forta_api_url()
    headers = get_forta_api_headers()
    query_options = LabelQueryOptions(dict)
    payload = query_options.get_query()

    response = requests.request(
        "POST", forta_api_url, json=payload, headers=headers)

    if response.status_code != 200:
        raise Exception(response.text)

    json_response = response.json()
    errors = json_response.get('errors')
    if errors is not None and len(errors) > 0:
        raise Exception(errors[0].get('message'))

    data = json_response.get('data')
    return LabelsResponse(data.get('labels'))


class LabelQueryOptions:
    def __init__(self, dict):
        self.entities = dict.get('entities')
        self.labels = dict.get('labels')
        self.sourceIds = dict.get('source_ids')
        self.entityType = dict.get('entity_type')
        self.state = dict.get('state')
        self.createdSince = dict.get('created_since')
        self.createdBefore = dict.get('created_before')
        self.first = dict.get('first')
        starting_cursor = dict.get('starting_cursor')
        if starting_cursor is not None:
            self.after = {
                'pageToken': starting_cursor.page_token if isinstance(starting_cursor, LabelCursor) else starting_cursor.get('page_token')
            }

    def get_query(self):
        query = """
          query($input: LabelsInput) {
            labels(input: $input) {
                labels {
                    createdAt
                    id
                    label {
                        confidence
                        entity
                        entityType
                        label
                        metadata
                        remove
                    }
                    source {
                        alertHash
                        alertId
                        bot {
                            id
                            image
                            imageHash
                            manifest
                        }
                        chainId
                        id
                    }
                }
                pageInfo {
                    hasNextPage
                    endCursor {
                        pageToken
                    }
                }
            }
          }
          """
        query_variables = vars(self)
        filtered_query_variables = {
            k: v for k, v in query_variables.items() if v is not None}
        return dict(query=query, variables={"input": filtered_query_variables})


class LabelsResponse:
    def __init__(self, dict):
        self.page_info = LabelPageInfo(dict.get('pageInfo')) if dict.get(
            'pageInfo') is not None else None
        self.labels = []
        labels_data = dict.get('labels', [])
        for label_data in labels_data:
            label_dict = label_data.get('label')
            label_dict['id'] = label_data.get('id')
            label_dict['source'] = label_data.get('source')
            label_dict['createdAt'] = label_data.get('createdAt')
            self.labels.append(Label(label_dict))


class LabelPageInfo:
    def __init__(self, dict):
        self.has_next_page = dict.get('hasNextPage')
        self.end_cursor = LabelCursor(dict.get('endCursor')) if dict.get(
            'endCursor') is not None else None


class LabelCursor:
    def __init__(self, dict):
        self.page_token = dict.get('pageToken')



import requests
import json
from .alert import Alert
from .finding import FindingType, FindingSeverity
from .label import EntityType
from .utils import get_forta_api_headers, get_forta_api_url


def send_alerts(alerts):
    if not isinstance(alerts, list):
        alerts = [alerts]

    alerts_api_url = get_forta_api_url()
    headers = get_forta_api_headers()
    mutation = SendAlertsRequest(alerts).get_mutation()

    response = requests.request(
        "POST", alerts_api_url, json=mutation, headers=headers)

    if response.status_code == 200:
        data = response.json().get('data')
        if data:
            return data.get('sendAlerts').get('alerts')
    else:
        message = response.text
        raise Exception(message)


class SendAlertsRequest:
    def __init__(self, alerts):
        self.alerts = []
        # serialize the alerts list
        for alert in alerts:
            # convert finding timestamp to RFC3339 format
            alert["finding"].timestamp = alert["finding"].timestamp.astimezone(
            ).isoformat()
            # serialize finding
            finding = json.loads(alert["finding"].toJson())
            # convert enums to all caps to match graphql enums
            finding["type"] = FindingType(finding["type"]).name.upper()
            finding["severity"] = FindingSeverity(
                finding["severity"]).name.upper()
            for label in finding.get("labels", []):
                label["entityType"] = EntityType(
                    label["entityType"]).name.upper()
            # remove protocol field (not part of graphql schema)
            del finding["protocol"]
            # remove any empty-value or snake-case-keyed fields
            finding = {k: v for k, v in finding.items()
                       if v is not None and "_" not in k}
            for index, label in enumerate(finding.get("labels", [])):
                finding["labels"][index] = {k: v for k, v in label.items()
                                            if v is not None and "_" not in k}
            self.alerts.append({
                "botId": alert["bot_id"],
                "finding": finding
            })

    def get_mutation(self):
        mutation = """
        mutation SendAlerts($alerts: [AlertRequestInput!]!) {
            sendAlerts(alerts: $alerts) {
                alerts {
                    alertHash
                    error {
                        code
                        message
                    }
                }
            }
        }
        """
        return dict(query=mutation, variables={"alerts": self.alerts})


def get_alerts(dict):
    alerts_api_url = get_forta_api_url()
    headers = get_forta_api_headers()
    query_options = AlertQueryOptions(dict)
    payload = query_options.get_query()

    response = requests.request(
        "POST", alerts_api_url, json=payload, headers=headers)

    if response.status_code == 200:
        data = response.json().get('data')
        if data:
            return AlertsResponse(data.get('alerts'))
    else:
        message = response.text
        raise Exception(message)


class AlertCursor:
    def __init__(self, dict):
        self.alert_id = dict.get('alertId')
        self.block_number = dict.get('blockNumber')


class AlertQueryOptions:
    def __init__(self, dict):
        self.bots = dict.get('bot_ids')
        self.addresses = dict.get('addresses')
        self.alertHash = dict.get('alert_hash')
        self.alertId = dict.get('alert_id')
        self.alertIds = dict.get('alert_ids')
        self.chainId = dict.get('chain_id')
        self.createdSince = dict.get('created_since')
        self.createdBefore = dict.get('created_before')
        self.first = dict.get('first')
        self.after = dict.get('starting_cursor')
        self.projectId = dict.get('project_id')
        self.scanNodeConfirmations = dict.get('scan_node_confirmations')
        self.severities = dict.get('severities')
        self.transactionHash = dict.get('transaction_hash')
        self.blockSortDirection = dict.get('block_sort_direction')
        self.blockDateRange = dict.get('block_date_range')
        self.blockNumberRange = dict.get('block_number_range')

    def get_query(self):
        query = """
          query($input: AlertsInput) {
            alerts(input: $input) {
                alerts {
                    alertId
                    addresses
                    contracts {
                        address
                        name
                        projectId
                    }
                    createdAt
                    description
                    hash
                    metadata
                    name
                    projects {
                        id
                    }
                    protocol
                    scanNodeCount
                    severity
                    source {
                        transactionHash
                        bot {
                            chainIds
                            createdAt
                            description
                            developer
                            docReference
                            enabled
                            id
                            image
                            name
                            reference
                            repository
                            projects
                            scanNodes
                            version
                        }
                        block {
                            number
                            hash
                            timestamp
                            chainId
                        }
                        sourceAlert {
                            hash
                            botId
                            timestamp
                            chainId
                        }
                    }
                    alertDocumentType
                    findingType
                    relatedAlerts
                    chainId
                    labels {
                        label
                        confidence
                        entity
                        entityType
                        remove
                        metadata
                    }
                    addressBloomFilter {
                        bitset
                        k
                        m
                    }
                }
                pageInfo {
                    hasNextPage
                    endCursor {
                        blockNumber
                        alertId
                    }
                }
            }
          }
          """
        query_variables = vars(self)
        filtered_query_variables = {
            k: v for k, v in query_variables.items() if v is not None}
        return dict(query=query, variables={"input": filtered_query_variables})


class AlertsResponse:
    def __init__(self, dict):
        self.alerts = list(map(lambda t: Alert(t), dict.get(
            'alerts', []))) if dict.get('alerts') is not None else []
        self.page_info = PageInfo(dict.get('pageInfo')) if dict.get(
            'pageInfo') is not None else None


class PageInfo:
    def __init__(self, dict):
        self.has_next_page = dict.get('hasNextPage')
        self.end_cursor = AlertCursor(dict.get('endCursor')) if dict.get(
            'endCursor') is not None else None



import responses
import pytest
from .utils import get_forta_api_url, get_forta_api_headers
from .label import Label, EntityType
from .labels_api import get_labels

mock_label = {
    "createdAt": "2023-05-16T07:43:36.007844434Z",
    "id": "0x6f5f0cf5ac536bf50a2c06ba14a8b619cbf3e856da740f53800b03e4a8f97ed3",
    "label": {
        "confidence": 0.8,
        "entity": "0x432c4faab3e87f1dfcea9a34c3f6b92dbb3c6082",
        "entityType": "ADDRESS",
        "label": "scammer-eoa",
        "metadata": [
            "alert_ids=SCAM-DETECTOR-ADDRESS-POISONING",
            "chain_id=1",
            "threat_description_url=https://forta.org/attacks#address-poisoning"
        ],
        "remove": False
    },
    "source": {
        "alertHash": "0xee2cfd0eee7c5b7b67d550ad8043044221e9b6aa43d6e7604595c6dd18e254fa",
        "alertId": "SCAM-DETECTOR-ADDRESS-POISONING",
        "bot": {
            "id": "0x1d646c4045189991fdfd24a66b192a294158b839a6ec121d740474bdacb3ab23",
            "image": "disco.forta.network/bafybeic2fzqhgypi7i7wqhxjvedph74euymruszdu7ghch6ocxf6msa7z4@sha256:2feb8176efa065de85f0b199bba84c291bd606e6c22e6b6c41b60f928c3cb29e",
            "imageHash": "2feb8176efa065de85f0b199bba84c291bd606e6c22e6b6c41b60f928c3cb29e",
            "manifest": "QmSM4tqw3p1y1jQyC6CAWVDPvqviv8F1ZkvdQc9H3ENwcb"
        },
        "chainId": 1,
        "id": "0x1d646c4045189991fdfd24a66b192a294158b839a6ec121d740474bdacb3ab23"
    }
}


@responses.activate
def test_raises_exception_if_api_returns_error():
    with pytest.raises(Exception, match=r".*some error.*"):
        mock_response = responses.Response(
            url=get_forta_api_url(),
            headers=get_forta_api_headers(),
            method="POST",
            json={"errors": [{"message": "some error"}]},
            status=200,
        )
        responses.add(mock_response)

        get_labels({})


@responses.activate
def test_invokes_api_and_returns_parsed_labels():
    mock_response = responses.Response(
        url=get_forta_api_url(),
        headers=get_forta_api_headers(),
        method="POST",
        json={"data": {"labels": {"pageInfo": {
            "hasNextPage": False}, "labels": [mock_label]}}},
        status=200,
    )
    responses.add(mock_response)

    response = get_labels({
        "source_ids": ["0xbotId1", "0xbotId2"],
        "entity_type": "ADDRESS",
        "created_since": 1234,
        "state": True,
    })

    page_info = response.page_info
    labels = response.labels
    assert page_info.has_next_page == False
    assert len(labels) == 1
    label = labels[0]
    assert isinstance(label, Label)
    assert label.entity_type == EntityType.Address
    assert label.label == mock_label["label"]["label"]
    assert label.confidence == mock_label["label"]["confidence"]
    assert label.remove == mock_label["label"]["remove"]
    assert len(label.metadata) == 3
    assert label.metadata["threat_description_url"] == "https://forta.org/attacks#address-poisoning"
    assert label.id == mock_label["id"]
    assert label.created_at == mock_label["createdAt"]
    assert label.source.chain_id == mock_label["source"]["chainId"]
    assert label.source.alert_hash == mock_label["source"]["alertHash"]



from enum import IntEnum
from .utils import assert_enum_value_in_dict, assert_non_empty_string_in_dict


class EntityType(IntEnum):
    Unknown = 0
    Address = 1
    Transaction = 2
    Block = 3
    Url = 4


class Label:
    def __init__(self, dict):
        entityTypeVal = dict.get('entity_type', dict.get('entityType'))
        self.entity_type = EntityType[entityTypeVal.title()] if type(
            entityTypeVal) == str else EntityType(entityTypeVal)
        assert_enum_value_in_dict(self.__dict__, 'entity_type', EntityType)
        assert_non_empty_string_in_dict(dict, 'entity')
        self.entity = dict['entity']
        self.confidence = dict['confidence']
        self.label = dict['label']
        self.remove = dict.get('remove', False)
        self.metadata = dict.get('metadata') if dict.get(
            'metadata') is not None else {}
        # if metadata is array, convert to map
        if type(self.metadata) is list:
            self.metadata_array_to_map()
        self.id = dict.get('id')
        self.source = LabelSource(dict.get('source')) if dict.get(
            'source') is not None else None
        self.created_at = dict.get('createdAt', dict.get('created_at'))

    def toDict(self):
        d = dict(self.__dict__, **{
            'entityType': self.entity_type,
        })
        return {k: v for k, v in d.items() if v is not None}

    def metadata_array_to_map(self):
        # convert string array to string key/value map using first '=' character as separator
        # (label metadata is received as string array from API)
        metadata_map = {}
        for item in self.metadata:
            separator_index = item.find('=')
            key = item[0:separator_index]
            value = item[separator_index+1:len(item)]
            metadata_map[key] = value
        self.metadata = metadata_map


class LabelSource:
    def __init__(self, dict):
        self.alert_hash = dict.get('alertHash')
        self.alert_id = dict.get('alertId')
        self.bot = LabelSourceBot(dict.get('bot')) if dict.get(
            'bot') is not None else None
        self.chain_id = dict.get('chainId')
        self.id = dict.get('id')


class LabelSourceBot:
    def __init__(self, dict):
        self.id = dict.get('id')
        self.image = dict.get('image')
        self.image_hash = dict.get('imageHash')
        self.manifest = dict.get('manifest')
